// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef SIZEHEURISTIC_HPP
#define SIZEHEURISTIC_HPP

#include "OrderingHeuristic.hpp"
#include "Image.hpp"

class SizeHeuristic : public OrderingHeuristic
{
    public:

        SizeHeuristic();
        DlxHead*           selectDlx() const override;
        ImageListSet       selectFila(int nrp, int nrs, const gridbitfield_t occupancy, bool dlxPristine = false) const override;
        bool               isFixedOrder() const override;
        std::ostream&      print(std::ostream& os) const override;
        const std::string& getEntityName() const override;

        static int countFits(
                ImageListSet imageListSet,
                const int* remainingShapeList,
                int nrs,
                const gridbitfield_t occupancy,
                int minFits);

    private:
        const std::string entityName;
};


inline SizeHeuristic::SizeHeuristic()
    :
        entityName("g")
{
}


inline bool SizeHeuristic::isFixedOrder() const
{
    return false;
}


inline const std::string& SizeHeuristic::getEntityName() const
{
    return entityName;
}


inline std::ostream& SizeHeuristic::print(std::ostream& os) const
{
    return os << "s";
}


inline int SizeHeuristic::countFits(
        ImageListSet imageListSet,
        const int* remainingShapeList,
        int nrs,
        const gridbitfield_t occupancy,
        int minFits)
{
    int numFits = 0;
    for(int si = 0; si < nrs; ++si)
    {
        ImageList& imageList = imageListSet[remainingShapeList[si]];
        for(ImageList::iterator ii = imageList.begin(); ii != imageList.end(); ++ii)
        {
            if(!((*ii)->layoutMask & occupancy))
            {
                // Quit as soon as we equal or exceed previous minFits.  The
                // ">" case is included in case zero is passed in for minFit
                // which would otherwise cause this early out to fail and
                // force the routine to run to completion.
                //
                if(++numFits >= minFits)
                    return minFits+1;
            }
        }
    }
    return numFits;
}

#endif
